////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_uniform.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

static __thread unsigned long flags = 0;

void crnlp_uniform_init(crnlp_uniform_lock_t* lock)
{
    int i = 0;

    lock->head = 0;

    for (i=0; i< SCHEDSIZE; i++)
    {
        lock->enabled[i] = 0;
    }

    lock->enabled[lock->head] = 1;

    memset(lock->schedule, 0, SCHEDSIZE*sizeof(resource_mask_t));

    mcs_init(&(lock->sublock));

    lock->pending_requests = 0;
}

uint64_t *crnlp_uniform_lock(crnlp_uniform_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_node_t *curr_node, *self;
    uint8_t start_in_sched;
    mcs_node_t mcs_local;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    self = &(lock->requests[ind]);

    mcs_lock(&(lock->sublock), &mcs_local);

    // Determine which row in the table in which to place this request
    if (lock->pending_requests > 0)
    {
        start_in_sched = (lock->head + 1) % SCHEDSIZE;
        while ((lock->schedule[start_in_sched] & self->requested) != 0)
        {
            start_in_sched = (start_in_sched + 1) % SCHEDSIZE;
        }
    }
    else
    {
        start_in_sched = lock->head;
    }

    // Set the next value for this request
    self->next = (start_in_sched + 1) % SCHEDSIZE;

    // Update blocked for the next row, the count of pending requests,
    // and set the bits in this row corresponding to the resources
    // needed by this request
    lock->blocked[self->next] += 1;
    lock->pending_requests += 1;
    lock->schedule[start_in_sched] |= self->requested;

    mcs_unlock(&(lock->sublock), &mcs_local);

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    // Wait until satisfied
    while (lock->enabled[start_in_sched] != 1)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;
  
    return ret;
}

uint64_t *crnlp_uniform_unlock(crnlp_uniform_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_node_t *curr_node, *self;
    mcs_node_t mcs_local;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    self = &(lock->requests[ind]);

    mcs_lock(&(lock->sublock), &mcs_local);

    // Clear the bits in the table corresponding to the resources
    // that were needed for this request
    lock->schedule[lock->head] &= ~(self->requested);

    // Decrement blocked and the pending request count, as this
    // request is now finished
    lock->blocked[self->next] -= 1;
    lock->pending_requests -= 1;

    // If this was the last request blocking the next row, then
    // mark the next row as enabled
    if (lock->blocked[self->next] == 0)
    {
        lock->enabled[lock->head] = 0;
        lock->head = self->next;
        lock->enabled[lock->head] = 1;
    }

    mcs_unlock(&(lock->sublock), &mcs_local);

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

