////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include <sched.h>
#include <inttypes.h>
#ifdef __APPLE__
#include <sys/uio.h>
#else
#include <sys/io.h>
#endif
#include <sys/mman.h>

#include "nesting.h"
#include "crnlp_uniform_sgls.h"
#include "crnlp_uniform_slls.h"
#include "crnlp_general_sgls.h"
#include "crnlp_general_slls.h"
#include "ticket.h"
#include "mem.h"

#define GHZ PROC_SPEED

int nr_cpus;
int nr_requested;
int nr_resources;
int cs_length;
int cs_alpha;
int cs_beta;

/*
    Algorithms:
      4: U-C-RNLP with Static Global Lock Server
      5: U-C-RNLP with Static Local Lock Servers
      9: G-C-RNLP with Static Global Lock Server
     10: G-C-RNLP with Static Local Lock Servers
*/
int alg;

crnlp_uniform_sgls_lock_t crnlp_uniform_sgls;
crnlp_uniform_slls_lock_t crnlp_uniform_slls;
crnlp_general_sgls_lock_t crnlp_general_sgls;
crnlp_general_slls_lock_t crnlp_general_slls;

ticket_lock_t csvlock;
FILE *csvfile;

int shared_objects[sizeof(resource_mask_t)*8];

static volatile int counter = 0;

struct thread_context {
    int id;
    int* objects;
};

#ifndef __APPLE__
int migrate_thread_to_cpu(int target_cpu)
{
    cpu_set_t *cpu_set;
    size_t sz;
    int ret;

    cpu_set = CPU_ALLOC(nr_cpus);
    sz = CPU_ALLOC_SIZE(nr_cpus);
    CPU_ZERO_S(sz, cpu_set);
    CPU_SET_S(target_cpu, sz, cpu_set);

    ret = sched_setaffinity(0, sz, cpu_set);

    CPU_FREE(cpu_set);

    return ret;
}
#endif

resource_mask_t random_resources(int requested, int available)
{
    int i, j, tmp;
    bool flag;
    int choices[nr_resources];
    resource_mask_t retval = 0;
    for (i = 0; i < requested; i++)
    {
        do
        {
            tmp = (rand() % available);
            flag = true;

            for (j = 0; j < i; j++)
            {
                if(tmp == choices[j])
                    flag = false;
            }
        } while (!flag);

        choices[i] = tmp;
        retval |= (1l << tmp);
    }

    return retval;
}

float random_cs_length(int lmax, int alpha, int beta)
{
    if ((rand() % 100) <= alpha)
    {
        return (beta / 100.0f) * lmax;
    }
    else
    {
        return lmax;
    }
}

void* servicing_thread(void* args){
    int i,j;
    int r,s;
    int server_id;
    unsigned long one, tmp, tmp2;
    unsigned int other_server_bits;
    crnlp_uniform_slls_server_lock_t *server_lock_u;
    volatile crnlp_general_sgls_queue_t *queue, *queue2;
    volatile crnlp_general_slls_queue_t *queue_slls, *queue2_slls;
    volatile struct crnlp_general_sgls_node_t *node, *next_node;
    volatile struct crnlp_general_slls_node_t *node_slls, *next_node_slls;
    crnlp_general_slls_server_lock_t *server_lock_g;
    struct thread_context* ctx = (struct thread_context*) args;

#ifndef __APPLE__
    migrate_thread_to_cpu(CPU_MAP[ctx->id]);
#endif

    one = 1;

    if (alg == 10) // G-C-RNLP + SLLS
    {
        server_id = ctx->id / (QUEUE_LENGTH / NUM_SERVERS);
        server_lock_g = &(crnlp_general_slls.servers[server_id]);
        other_server_bits = 0;
    }
    else if (alg == 5) // U-C-RNLP + SLLS
    {
        server_id = ctx->id / (QUEUE_LENGTH / NUM_SERVERS);
        server_lock_u = &(crnlp_uniform_slls.servers[server_id]);
        other_server_bits = 0;
    }

    i = 1;
    while (1 == 1)
    {
        pthread_testcancel();

        switch (alg)
        {
        case 4: // U-C-RNLP + SGLS
            if (crnlp_uniform_sgls.requests[i].needs_service == 1)
            {
                crnlp_uniform_sgls_node_t *self = &(crnlp_uniform_sgls.requests[i]);

                if (crnlp_uniform_sgls.requests[i].done == 1)
                {
                    // Unlock code

                    // Clear the bits in the table corresponding to the resources
                    // that were needed for this request
                    crnlp_uniform_sgls.schedule[crnlp_uniform_sgls.head] &= ~(self->requested);

                    // Decrement blocked and the pending request count, as this
                    // request is now finished
                    crnlp_uniform_sgls.blocked[self->next] -= 1;
                    crnlp_uniform_sgls.pending_requests -= 1;

                    // If this was the last request blocking the next row, then
                    // mark the next row as enabled
                    if (crnlp_uniform_sgls.blocked[self->next] == 0)
                    {
                        crnlp_uniform_sgls.enabled[crnlp_uniform_sgls.head] = 0;
                        crnlp_uniform_sgls.head = self->next;
                        crnlp_uniform_sgls.enabled[crnlp_uniform_sgls.head] = 1;
                    }

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
                else
                {
                    // Lock code
                    uint8_t start_in_sched;

                    // Determine which row in the table in which to place
                    // this request
                    if (crnlp_uniform_sgls.pending_requests > 0)
                    {
                        start_in_sched = (crnlp_uniform_sgls.head + 1) % SCHEDSIZE;
                        while ((crnlp_uniform_sgls.schedule[start_in_sched] & self->requested) != 0)
                        {
                            start_in_sched = (start_in_sched + 1) % SCHEDSIZE;
                        }
                    }
                    else
                    {
                        start_in_sched = crnlp_uniform_sgls.head;
                    }

                    // Set the next value for this request
                    self->next = (start_in_sched + 1) % SCHEDSIZE;

                    // Update blocked for the next row, the count of pending requests,
                    // and set the bits in this row corresponding to the resources
                    // needed by this request
                    crnlp_uniform_sgls.blocked[self->next] += 1;
                    crnlp_uniform_sgls.pending_requests += 1;
                    crnlp_uniform_sgls.schedule[start_in_sched] |= self->requested;

                    // Set the spin location for this request to be
                    // its row in the table; it will be satisfied when
                    // enabled is set to true for this row
                    self->spin_location = start_in_sched;

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
            }

            break;
        case 5: // U-C-RNLP + SLLS
            if (server_lock_u->requests[i].needs_service == 1)
            {
                crnlp_uniform_slls_node_t *self = &(server_lock_u->requests[i]);
                if (server_lock_u->requests[i].done == 1)
                {
                    // Unlock code

                    // Clear the bits in the table corresponding to the resources
                    // that were needed for this request
                    server_lock_u->schedule[server_lock_u->head] &= ~(self->requested);

                    // Decrement blocked and the pending request count, as this
                    // request is now finished
                    server_lock_u->blocked[self->next] -= 1;
                    server_lock_u->pending_requests -= 1;

                    // If this was the last request blocking the next row, then
                    // update the head pointer and disable the row
                    if (server_lock_u->blocked[self->next] == 0)
                    {
                        server_lock_u->enabled[server_lock_u->head] = 0;
                        server_lock_u->head = self->next;
                    }

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;

                    if (server_lock_u->blocked[self->next] == 0)
                    {
                        // Clear the XBITS for this server, allowing another
                        // lock server to become enabled
                        if (server_id == 0)
                        {
                            __sync_fetch_and_and(&crnlp_uniform_slls.S, ~ABITS);
                        }
                        else
                        {
                            __sync_fetch_and_and(&crnlp_uniform_slls.S, ~BBITS);
                        }

                        // If there are still requests on this server, update
                        // the wave and present bits
                        if (server_lock_u->pending_requests > 0)
                        {
                            server_lock_u->rr_wave = ~(server_lock_u->rr_wave);
                            if (server_id == 0)
                            {
                                other_server_bits = __sync_fetch_and_add(
                                                        &crnlp_uniform_slls.S,
                                                        APRES | (server_lock_u->rr_wave & APHID));
                                other_server_bits = other_server_bits & BBITS;
                            }
                            else
                            {
                                other_server_bits = __sync_fetch_and_add(
                                                        &crnlp_uniform_slls.S,
                                                        BPRES | (server_lock_u->rr_wave & BPHID));
                                other_server_bits = other_server_bits & ABITS;
                            }
                        }
                    }
                }
                else
                {
                    // Lock code
                    uint8_t start_in_sched;

                    // Determine which row in the table in which to place
                    // this request
                    if (server_lock_u->pending_requests > 0)
                    {
                        start_in_sched = (server_lock_u->head + 1) % SCHEDSIZE;
                        while ((server_lock_u->schedule[start_in_sched] & self->requested) != 0)
                        {
                            start_in_sched = (start_in_sched + 1) % SCHEDSIZE;
                        }
                    }
                    else
                    {
                        start_in_sched = server_lock_u->head;
                    }

                    // Set the next value for this request
                    self->next = (start_in_sched + 1) % SCHEDSIZE;

                    // Update blocked for the next row, the count of pending requests,
                    // and set the bits in this row corresponding to the resources
                    // needed by this request
                    server_lock_u->blocked[self->next] += 1;
                    server_lock_u->pending_requests += 1;
                    server_lock_u->schedule[start_in_sched] |= self->requested;

                    // Set the spin location for this request to be
                    // its row in the table; it will be satisfied when
                    // enabled is set to true for this row
                    self->spin_location = start_in_sched;

                    mb();

                    // If the server was idle, update the phase and mark the
                    // server present
                    if (server_lock_u->pending_requests == 1)
                    {
                        server_lock_u->rr_wave = ~(server_lock_u->rr_wave);
                        if (server_id == 0)
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &crnlp_uniform_slls.S,
                                                    APRES | (server_lock_u->rr_wave & APHID));
                            other_server_bits = other_server_bits & BBITS;
                        }
                        else
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &crnlp_uniform_slls.S,
                                                    BPRES | (server_lock_u->rr_wave & BPHID));
                            other_server_bits = other_server_bits & ABITS;
                        }
                    }

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
            }

            // If the head is not enabled, check the RR logic to determine if
            // this lock server should be enabled
            if ((server_lock_u->pending_requests > 0) &&
                (server_lock_u->enabled[server_lock_u->head] != 1))
            {
                if ((server_id == 0) &&
                    ((other_server_bits == 0) ||
                     (other_server_bits != (crnlp_uniform_slls.S & BBITS))))
                {
                    server_lock_u->enabled[server_lock_u->head] = 1;
                }
                else if ((server_id == 1) &&
                    ((other_server_bits == 0) ||
                     (other_server_bits != (crnlp_uniform_slls.S & ABITS))))
                {
                    server_lock_u->enabled[server_lock_u->head] = 1;
                }
            }

            break;
        case 9: // G-C-RNLP + SGLS
            if (crnlp_general_sgls.requests[i].needs_service == 1)
            {
                crnlp_general_sgls_request_t *self = &(crnlp_general_sgls.requests[i]);

                if (crnlp_general_sgls.requests[i].done == 1)
                {
                    // Unlock code

                    // Iterate through the resource queues for this request
                    tmp = self->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        queue = &(crnlp_general_sgls.queues[r]);
                        node = queue->head;

                        // Update the head
                        queue->head = queue->head->next;

                        // Update the new head's prev pointer
                        if (queue->head)
                        {
                            queue->head->prev = NULL;
                        }
                        else
                        {
                            queue->tail = NULL;
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    // Check the heads of the resource queues to determine
                    // if they are now enabled
                    tmp = self->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        queue = &(crnlp_general_sgls.queues[r]);

                        // If the queue is empty, skip it
                        if (queue->head)
                        {
                            // Check if the head of this queue is now enabled,
                            // by checking the queues for this request
                            int is_enabled = 1;

                            tmp2 = queue->head->request->requested;
                            s = ffsl(tmp2);
                            while (s != 0)
                            {
                                s = s - 1;

                                queue2 = &(crnlp_general_sgls.queues[s]);

                                // If this request is not the head, it can't
                                // be enabled
                                if (queue->head->request != queue2->head->request)
                                {
                                    is_enabled = 0;
                                    break;
                                }

                                tmp2 &= ~(one << s);
                                s = ffsl(tmp2);
                            }

                            // If the request at the head is enabled, set
                            // its bit
                            if (is_enabled == 1)
                            {
                                queue->head->request->enabled = 1;
                            }
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    // Decrement the pending request count, as this
                    // request is now finished
                    crnlp_general_sgls.pending_requests -= 1;

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
                else
                {
                    // Lock code
                    uint64_t start_in_sched, prev_start, potential_start;

                    // Find the earliest possible start time
                    start_in_sched = get_cycles();
                    prev_start = start_in_sched;

                    while (1)
                    {
                        // Loop through the required resources
                        tmp = self->requested;
                        r = ffsl(tmp);
                        while (r != 0)
                        {
                            r = r - 1;

                            // For this resource, find the earliest possible start time after
                            // the current start_in_sched within the resource queue
                            queue = &(crnlp_general_sgls.queues[r]);

                            if (queue->head)
                            {
                                // If this request cannot be satisfied before the first node,
                                // we need to find the node it will follow
                                if ((queue->head->request->enabled == 1) ||
                                    (get_cycles() + self->length >= queue->head->start_time))
                                {
                                    node = queue->head->next;

                                    while (node)
                                    {
                                        potential_start = node->prev->start_time +
                                                          node->prev->request->length;
                                        potential_start = potential_start < start_in_sched ?
                                                              start_in_sched :
                                                              potential_start;
                                        if (potential_start + self->length < node->start_time)
                                        {
                                            // This location is delay-preserving!
                                            start_in_sched = potential_start;
                                            break;
                                        }
                                        else
                                        {
                                            // We need to continue looking in the schedule
                                            node = node->next;
                                        }
                                    }

                                    // If node is NULL, we belong at the tail of the queue
                                    if (!node)
                                    {
                                        potential_start = queue->tail->start_time +
                                                          queue->tail->request->length;
                                        if (potential_start > start_in_sched)
                                        {
                                            start_in_sched = potential_start;
                                        }
                                    }
                                }
                            }

                            tmp &= ~(one << r);
                            r = ffsl(tmp);
                        }

                        if (prev_start == start_in_sched)
                        {
                            // If the selected time did not change, we are done looping
                            break;
                        }
                        else
                        {
                            // Otherwise, update prev_start and go around again
                            prev_start = start_in_sched;
                        }
                    }

                    // Update pointers to insert the request into the queue for each
                    // required resource
                    j = 0;
                    int is_head = 1;
                    tmp = self->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        // Grab a queue from the pool
                        node = &(crnlp_general_sgls.nodes[i][j++]);

                        node->request = self;
                        node->start_time = start_in_sched;

                        node->prev = NULL;
                        node->next = NULL;

                        // Insert this node into the queue
                        queue = &(crnlp_general_sgls.queues[r]);
                        if (!queue->head)
                        {
                            queue->head = node;
                            queue->tail = node;
                        }
                        else
                        {
                            // Find the node it will precede
                            next_node = queue->head;
                            while (next_node && (next_node->start_time <= start_in_sched))
                            {
                                next_node = next_node->next;
                            }

                            // Insert the new node before next_node
                            if (!next_node)
                            {
                                queue->tail->next = node;
                                node->prev = queue->tail;

                                queue->tail = node;

                                is_head = 0;
                            }
                            else if (next_node == queue->head)
                            {
                                queue->head = node;

                                next_node->prev = node;
                                node->next = next_node;
                            }
                            else
                            {
                                next_node->prev->next = node;
                                node->prev = next_node->prev;

                                next_node->prev = node;
                                node->next = next_node;

                                is_head = 0;
                            }
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    if (is_head == 1)
                    {
                        self->enabled = 1;
                    }

                    // Update the count of pending requests
                    crnlp_general_sgls.pending_requests += 1;

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
            }

            break;
        case 10: // G-C-RNLP + SLLS
            if (server_lock_g->requests[i].needs_service == 1)
            {
                crnlp_general_slls_request_t *self = &(server_lock_g->requests[i]);

                if (server_lock_g->requests[i].done == 1)
                {
                    // Unlock code

                    // Iterate through the resource queues for this request
                    tmp = self->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        queue_slls = &(server_lock_g->queues[r]);
                        node_slls = queue_slls->head;

                        // Update the head
                        queue_slls->head = queue_slls->head->next;

                        // Update the new head's prev pointer
                        if (queue_slls->head)
                        {
                            queue_slls->head->prev = NULL;
                        }
                        else
                        {
                            queue_slls->tail = NULL;
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    // Decrement the pending and satisfied request counts,
                    // as this request is now finished
                    server_lock_g->pending_requests -= 1;
                    server_lock_g->satisfied_requests -= 1;

                    mb();

                    // Mark this request as serviced
                    self->needs_service = 0;

                    if (server_lock_g->satisfied_requests == 0)
                    {
                        server_lock_g->enabled = 0;

                        // Clear the XBITS for this server, allowing another
                        // server to become enabled
                        if (server_id == 0)
                        {
                            __sync_fetch_and_and(&crnlp_general_slls.S, ~ABITS);
                        }
                        else
                        {
                            __sync_fetch_and_and(&crnlp_general_slls.S, ~BBITS);
                        }

                        // If there are still requests on this server, update
                        // the wave and present bits
                        if (server_lock_g->pending_requests > 0)
                        {
                            server_lock_g->rr_wave = ~(server_lock_g->rr_wave);
                            if (server_id == 0)
                            {
                                other_server_bits = __sync_fetch_and_add(
                                                        &crnlp_general_slls.S,
                                                        APRES | (server_lock_g->rr_wave & APHID));
                                other_server_bits = other_server_bits & BBITS;
                            }
                            else
                            {
                                other_server_bits = __sync_fetch_and_add(
                                                        &crnlp_general_slls.S,
                                                        BPRES | (server_lock_g->rr_wave & BPHID));
                                other_server_bits = other_server_bits & ABITS;
                            }
                        }
                    }
                }
                else
                {
                    // Lock code
                    uint64_t start_in_sched, prev_start, potential_start;

                    // Find the earliest possible start time
                    start_in_sched = get_cycles();
                    prev_start = start_in_sched;

                    while (1)
                    {
                        // Loop through the required resources
                        tmp = self->requested;
                        r = ffsl(tmp);
                        while (r != 0)
                        {
                            r = r - 1;

                            // For this resource, find the earliest possible start time after
                            // the current start_in_sched within the resource queue
                            queue_slls = &(server_lock_g->queues[r]);

                            if (queue_slls->head)
                            {
                                // If this request cannot be satisfied before the first node,
                                // we need to find the node it will follow
                                if ((queue_slls->head->request->enabled == 1) ||
                                    (get_cycles() + self->length >= queue_slls->head->start_time))
                                {
                                    node_slls = queue_slls->head->next;

                                    while (node_slls)
                                    {
                                        potential_start = node_slls->prev->start_time +
                                                          node_slls->prev->request->length;
                                        potential_start = potential_start < start_in_sched ?
                                                              start_in_sched :
                                                              potential_start;
                                        if (potential_start + self->length < node_slls->start_time)
                                        {
                                            // This location is delay-preserving!
                                            start_in_sched = potential_start;
                                            break;
                                        }
                                        else
                                        {
                                            // We need to continue looking in the schedule
                                            node_slls = node_slls->next;
                                        }
                                    }

                                    // If node is NULL, we belong at the tail of the queue
                                    if (!node_slls)
                                    {
                                        potential_start = queue_slls->tail->start_time +
                                                          queue_slls->tail->request->length;
                                        if (potential_start > start_in_sched)
                                        {
                                            start_in_sched = potential_start;
                                        }
                                    }
                                }
                            }

                            tmp &= ~(one << r);
                            r = ffsl(tmp);
                        }

                        if (prev_start == start_in_sched)
                        {
                            // If the selected time did not change, we are done looping
                            break;
                        }
                        else
                        {
                            // Otherwise, update prev_start and go around again
                            prev_start = start_in_sched;
                        }
                    }

                    // Update pointers to insert the request into the queue for each
                    // required resource
                    j = 0;
                    int is_head = 1;
                    tmp = self->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        // Grab a queue node from the pool
                        node_slls = &(server_lock_g->nodes[i][j++]);

                        node_slls->request = self;
                        node_slls->start_time = start_in_sched;

                        node_slls->prev = NULL;
                        node_slls->next = NULL;

                        // Insert this node into the queue
                        queue_slls = &(server_lock_g->queues[r]);
                        if (!queue_slls->head)
                        {
                            queue_slls->head = node_slls;
                            queue_slls->tail = node_slls;
                        }
                        else
                        {
                            // Find the node it will precede
                            next_node_slls = queue_slls->head;
                            while (next_node_slls && (next_node_slls->start_time <= start_in_sched))
                            {
                                next_node_slls = next_node_slls->next;
                            }

                            // Insert the new node before next_node
                            if (!next_node_slls)
                            {
                                queue_slls->tail->next = node_slls;
                                node_slls->prev = queue_slls->tail;

                                queue_slls->tail = node_slls;

                                is_head = 0;
                            }
                            else if (next_node_slls == queue_slls->head)
                            {
                                queue_slls->head = node_slls;

                                next_node_slls->prev = node_slls;
                                node_slls->next = next_node_slls;
                            }
                            else
                            {
                                next_node_slls->prev->next = node_slls;
                                node_slls->prev = next_node_slls->prev;

                                next_node_slls->prev = node_slls;
                                node_slls->next = next_node_slls;

                                is_head = 0;
                            }
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    // Update the count of pending requests
                    server_lock_g->pending_requests += 1;

                    mb();

                    // If the server was idle, update the phase and mark the
                    // server present
                    if (server_lock_g->pending_requests == 1)
                    {
                        server_lock_g->rr_wave = ~(server_lock_g->rr_wave);
                        if (server_id == 0)
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &crnlp_general_slls.S,
                                                    APRES | (server_lock_g->rr_wave & APHID));
                            other_server_bits = other_server_bits & BBITS;
                        }
                        else
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &crnlp_general_slls.S,
                                                    BPRES | (server_lock_g->rr_wave & BPHID));
                            other_server_bits = other_server_bits & ABITS;
                        }
                    }

                    // Mark this request as serviced
                    self->needs_service = 0;
                }
            }

            // If is server is not enabled, check the RR logic to determine
            // if it should be enabled
            if ((server_lock_g->pending_requests > 0) &&
                (server_lock_g->enabled != 1))
            {
                uint8_t should_enable = 0;
                if ((server_id == 0) &&
                    ((other_server_bits == 0) ||
                     (other_server_bits != (crnlp_general_slls.S & BBITS))))
                {
                    should_enable = 1;
                }
                else if ((server_id == 1) &&
                    ((other_server_bits == 0) ||
                     (other_server_bits != (crnlp_general_slls.S & ABITS))))
                {
                    should_enable = 1;
                }

                if (should_enable == 1)
                {
                    // Mark the server as enabled
                    server_lock_g->enabled = 1;

                    // Check the heads of the resource queues to determine
                    // if they are now enabled
                    for (r = 0; r < RES_COUNT; r++)
                    {
                        queue_slls = &(server_lock_g->queues[r]);

                        // If the queue is empty or the request at the head
                        // is enabled, skip it
                        if (queue_slls->head && queue_slls->head->request->enabled == 0)
                        {
                            // Check if the head of this queue is now enabled,
                            // by checking the queues for this request
                            int is_enabled = 1;

                            tmp2 = queue_slls->head->request->requested;
                            s = ffsl(tmp2);
                            while (s != 0)
                            {
                                s = s - 1;

                                queue2_slls = &(server_lock_g->queues[s]);

                                // If this request is not the head, it can't
                                // be enabled
                                if (queue_slls->head->request != queue2_slls->head->request)
                                {
                                    is_enabled = 0;
                                    break;
                                }

                                tmp2 &= ~(one << s);
                                s = ffsl(tmp2);
                            }

                            // If the request at the head is enabled, set
                            // its bit
                            if (is_enabled == 1)
                            {
                                server_lock_g->satisfied_requests++;

                                queue_slls->head->request->enabled = 1;
                            }
                        }
                    }
                }
            }

            break;
        }

        // Choose the next request to check for needing service
        if (alg == 4 || alg == 9) // global LS
        {
            // Move onto the next request (skip core 0, the lock server)
            i = (i+1) % nr_cpus;
            if (i == 0) { i=1; }
        }
        else // local LS
        {
            // Move onto the next request (skip core 0, the lock server)
            if (nr_cpus > CORES_PER_SOCKET)
            {
                if (server_id == 0)
                {
                    i = (i+1) % CORES_PER_SOCKET;
                    if (i == 0) { i=1; }
                }
                else
                {
                    i = (i+1) % nr_cpus;
                    if (i == 0) { i=CORES_PER_SOCKET+1; }
                }
            }
            else
            {
                i = (i+1) % nr_cpus;
                if (i == 0) { i=1; }
            }
        }
    }
}

void* rt_thread(void* args)
{
    crnlp_uniform_sgls_node_t *crnlp_uniform_sgls_self;
    crnlp_uniform_slls_node_t *crnlp_uniform_slls_self;
    crnlp_general_sgls_request_t *crnlp_general_sgls_self;
    crnlp_general_slls_request_t *crnlp_general_slls_self;
    int server_id;
    int i,j,k,l,r;
    struct thread_context* ctx = (struct thread_context*) args;
    resource_mask_t tmp;
    uint64_t max_l = 0, max_u = 0;
    uint64_t sum_l = 0, sum_u = 0;
    uint64_t lock[BUFSIZE], unlock[BUFSIZE];
    uint64_t block[BUFSIZE];
    uint64_t end_time;
    resource_mask_t requests[TRIALS];
    float cs_lengths[TRIALS];
    uint64_t *ret_tmp = NULL;

#ifndef __APPLE__
    migrate_thread_to_cpu(CPU_MAP[ctx->id]);
#endif

    // Determine resources at init time so as not to alter the runtime demand
    // for the lock
    for (i = 0; i < TRIALS; i++)
    {
        requests[i] = random_resources(nr_requested, nr_resources);

        cs_lengths[i] = random_cs_length(cs_length, cs_alpha, cs_beta);
    }

    crnlp_uniform_sgls_self = &(crnlp_uniform_sgls.requests[ctx->id]);
    crnlp_general_sgls_self = &(crnlp_general_sgls.requests[ctx->id]);

    server_id = ctx->id / (QUEUE_LENGTH / NUM_SERVERS);
    crnlp_uniform_slls_self = &(crnlp_uniform_slls.servers[server_id].requests[ctx->id]);
    crnlp_general_slls_self = &(crnlp_general_slls.servers[server_id].requests[ctx->id]);

    for (l = 0; l < TRIALS/BUFSIZE; l++)
    {
        __sync_fetch_and_add(&counter, 1);

        if (alg == 4 || alg == 9) // global LS
        {
            while (counter != (nr_cpus-1)*(l+1))
            {
                cpu_relax();
            }
        }
        else // local LS
        {
            while (counter != ((nr_cpus > CORES_PER_SOCKET ? nr_cpus-2 : nr_cpus-1))*(l+1))
            {
                cpu_relax();
            }
        }
        mb();
    
        for (i = 0; i < BUFSIZE; i++)
        {
            k = l*BUFSIZE + i;

            switch (alg)
            {
            case 4: // U-C-RNLP + SGLS
                crnlp_uniform_sgls_self->requested = requests[k];
                crnlp_uniform_sgls_self->next = -1;
                crnlp_uniform_sgls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_uniform_sgls_lock(&crnlp_uniform_sgls, ctx->id);

                break;
            case 5: // U-C-RNLP + SLLS
                crnlp_uniform_slls_self->requested = requests[k];
                crnlp_uniform_slls_self->next = -1;
                crnlp_uniform_slls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_uniform_slls_lock(&crnlp_uniform_slls, ctx->id);

                break;
            case 9: // G-C-RNLP + SGLS
                crnlp_general_sgls_self->requested = requests[k];
                crnlp_general_sgls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_general_sgls_lock(&crnlp_general_sgls, ctx->id);

                break;
            case 10: // G-C-RNLP + SLLS
                crnlp_general_slls_self->requested = requests[k];
                crnlp_general_slls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_general_slls_lock(&crnlp_general_slls, ctx->id);

                break;
            }

            lock[i] = ret_tmp[0];
            block[i] = ret_tmp[1];

            free(ret_tmp);
            ret_tmp = NULL;

            /*******code for critical section*******/
            end_time = get_cycles() + (int)(GHZ * cs_lengths[k]); // * 1000 should be 1ms
            while(get_cycles() < end_time){
              cpu_relax();
            }
            /***************************************/
            __sync_synchronize();

            switch (alg)
            {
            case 4: // U-C-RNLP + SGLS
                ret_tmp = crnlp_uniform_sgls_unlock(&crnlp_uniform_sgls, ctx->id);

                break;
            case 5: // U-C-RNLP + SLLS
                ret_tmp = crnlp_uniform_slls_unlock(&crnlp_uniform_slls, ctx->id);

                break;
            case 9: // G-C-RNLP + SGLS
                ret_tmp = crnlp_general_sgls_unlock(&crnlp_general_sgls, ctx->id);

                break;
            case 10: // G-C-RNLP + SLLS
                ret_tmp = crnlp_general_slls_unlock(&crnlp_general_slls, ctx->id);

                break;
            }

            unlock[i] = ret_tmp[0];

            free(ret_tmp);
            ret_tmp = NULL;
        }

        ticket_lock(&csvlock);

        for (i = 0; i < BUFSIZE; i++)
        {
            k = l*BUFSIZE + i;

            fprintf(csvfile, "%d, %d, %d, %d, %" PRIu64 ", %" PRIu64 ", %" PRIu64", %lu, %f\n",
                    ctx->id, nr_cpus, nr_resources, nr_requested,
                    lock[i], unlock[i], block[i],
                    requests[k], cs_lengths[k]);

            if(lock[i] > max_l)
                max_l = lock[i];

            if(unlock[i] > max_u)
                max_u = unlock[i];
           
            sum_l += lock[i];
            sum_u += unlock[i];
        }

        ticket_unlock(&csvlock);
    }

    //fprintf(stderr,"Thread %d: max lock %f, max unlock %f\navg lock %f, avg unlock %f\n",
    //        ctx->id, max_l/GHZ, max_u/GHZ, sum_l/(TRIALS*GHZ), sum_u/(TRIALS*GHZ));

    return NULL;
}

int main(int argc, char** argv)
{
    if (argc != 9)
    {
        printf("argc: %d\n",argc);
        fprintf(stderr, "Usage: run cpus nr_requested nr_resources cs_length alg csvfile\n");
        return 1;
    }

    int i;
    pthread_t *tasks;
    struct thread_context* ctx;
    
    mlockall(MCL_CURRENT | MCL_FUTURE);

    nr_cpus = atoi(argv[1]);
    nr_requested = atoi(argv[2]);
    nr_resources = atoi(argv[3]);
    cs_length = atoi(argv[4]);
    cs_alpha = atoi(argv[5]); // fraction of requests to be shorter, as int in [0,100]
    cs_beta = atoi(argv[6]);  // fraction of L_max for short requests, as int in [0,100]
    alg = atoi(argv[7]);
    csvfile = fopen(argv[8], "w");

    tasks = calloc(nr_cpus, sizeof(pthread_t));
    ctx = malloc(nr_cpus * sizeof(struct thread_context));

    fprintf(csvfile, "thread, numcores, numresources, numrequested, lock, unlock, block, resources\n");

    crnlp_uniform_sgls_init(&crnlp_uniform_sgls);
    crnlp_uniform_slls_init(&crnlp_uniform_slls);
    crnlp_general_sgls_init(&crnlp_general_sgls);
    crnlp_general_slls_init(&crnlp_general_slls);
    ticket_init(&csvlock);

#ifdef DISABLE_INTERRUPTS
    iopl(3);
#endif

    for(i = 0; i < nr_cpus; i++)
    {
        ctx[i].id = i;
        ctx[i].objects = shared_objects;
    }

    mb();

    if (alg == 4 || alg == 9) // global LS
    {
        // Servicing thread
        pthread_create(&tasks[0], NULL, servicing_thread, (void *) (&ctx[0]));

        // Remaining m-1 request threads
        for(i = 1; i < nr_cpus; i++)
        {
            mb();
            pthread_create(&tasks[i], NULL, rt_thread, (void *) (&ctx[i]));
            mb();
        }

        // Wait on the request threads
        for(i = 1; i < nr_cpus; i++)
        {
            pthread_join(tasks[i], NULL);
        }

        // Kill the servicing thread
        pthread_cancel(tasks[0]);
        pthread_join(tasks[0], NULL);
    }
    else // local LS
    {
        // Servicing threads
        pthread_create(&tasks[0], NULL, servicing_thread, (void *) (&ctx[0]));
        if (nr_cpus > CORES_PER_SOCKET)
        {
            pthread_create(&tasks[CORES_PER_SOCKET], NULL, servicing_thread, (void *) (&ctx[CORES_PER_SOCKET]));
        }

        // Remaining m-1 request threads
        for(i = 0; i < nr_cpus; i++)
        {
            if (i == 0 || i == CORES_PER_SOCKET) { continue; }
            mb();
            pthread_create(&tasks[i], NULL, rt_thread, (void *) (&ctx[i]));
            mb();
        }

        // Wait on the request threads
        for(i = 0; i < nr_cpus; i++)
        {
            if (i == 0 || i == CORES_PER_SOCKET) { continue; }
            pthread_join(tasks[i], NULL);
        }

        // Kill the servicing threads
        pthread_cancel(tasks[0]);
        pthread_join(tasks[0], NULL);
        if (nr_cpus > CORES_PER_SOCKET)
        {
            pthread_cancel(tasks[CORES_PER_SOCKET]);
            pthread_join(tasks[CORES_PER_SOCKET], NULL);
        }
    }

    return 0;
}
